"""
Functionality for interaction with factory data
on the FactoryMind platform

Code documentation
^^^^^^^^^^^^^^^^^^
"""

# Standard library imports
# Third party imports
import pandas as pd
import requests

from factorymind.exceptions import (
    CouldNotFetchDataException,
    DatasetDoesNotExistException,
)

# API_URL_BASE = "core-api"


class FactoryDB(object):
    """Class for FM interaction with factory data

    Class arguments
    ---------------
    :param api_url_base: str, defaults to "core-api:80".
        Url to FM api
    :param api_key: str, default `None`.
        API key to use with FM API
    :param verbose: boolean, default `True`
        Indicator to print out list of data sources
    """

    def __init__(
        self,
        api_url_base: str = "core-api:80",
        api_key: str = None,
        verbose: bool = True,
    ):
        """Constructor method"""
        self.api_url_base = api_url_base
        self.api_key = api_key
        self.verbose = verbose
        self.datasets = self.list_data_sources()

    def list_data_sources(self):
        """List available data sources on factory data platform

        Arguments:
        ----------

        Returns:
        --------
        :return data_sources: list of strings.
            List of data sources (collections) available on FactoryMind data platform
        :rtype: list

        Example
        ^^^^^^^
        .. code-block:: python

            >>> from factorymind.data_loader import FactoryDB

            >>> mydb = FactoryDB(apikey=YOUR-API-KEY)
            >>> mydb.list_data_sources()

            ['example_data.energy_demand', 'example_data.sensors', 'sensors.sensors', 'sensors.sensors_metadata']
        """
        resp = requests.get(
            f"http://{self.api_url_base}/list_data_sources?verbose=true"
        )
        data_sources = self._get_response(resp)
        return data_sources

    def data_info(self, dataset: str):
        """More info on specific dataset.

        Arguments:
        ----------
        :param dataset: str.
            Name of dataset (table) to get info about.

        Returns:
        --------
        :return dataset_info: dict.
            Info of dataset, keys "name", "nrows", "ncolumns"
        :return dataset_columns: list.
            List of dataset columns (strings)
        :return dataset_sample: pd.DataFrame.
            Sample of 5 last records in dataset

        Example
        ^^^^^^^
        .. code-block:: python

            >>> from factorymind.data_loader import FactoryDB

            >>> mydb = FactoryDB(apikey=YOUR-API-KEY)
            >>> table_info, _, df_sample = mydb.data_info(dataset='example_data.energy_demand')
            >>> print(table_info)

            {'name': 'example_data.energy_demand', 'nrows': 35064, 'ncolumns': 29}
        """
        self._check_if_dataset_exists(dataset)
        resp = requests.get(f"http://{self.api_url_base}/data_info?dataset={dataset}")
        data_resp = self._get_response(resp)

        # Construct output
        dataset_info = {key: data_resp[key] for key in ["name", "nrows", "ncolumns"]}
        dataset_columns = data_resp["columns"]
        dataset_sample = pd.DataFrame(data_resp["sample"], columns=dataset_columns)

        return dataset_info, dataset_columns, dataset_sample

    def get_data(self, dataset: str, limit: int = 10000):
        """Fetch batch dataset from DB

        Arguments:
        ----------
        :param dataset: str.
            Name of dataset (collection)
        :param limit: int, default 10000.
            Number of datapoints (rows) to limit query to

        Returns:
        --------
        :return data: pd.DataFrame.
            Dataset, limited to last "limit" records

        Example
        ^^^^^^^
        .. code-block:: python

            >>> from factorymind.data_loader import FactoryDB

            >>> mydb = FactoryDB(apikey=YOUR-API-KEY)
            >>> df = mydb.get_data(dataset="example_data.sensors", limit=100)
        """
        self._check_if_dataset_exists(dataset)
        resp = requests.get(
            f"http://{self.api_url_base}/get_data?dataset={dataset}&limit={limit}"
        )
        data_resp = self._get_response(resp)

        # Construct output
        if "columns" in data_resp.keys():
            data = pd.DataFrame(data_resp["data"], columns=data_resp["columns"])
        else:
            data = pd.DataFrame(data_resp["data"])

        return data

    def run_custom_query(self, query: str):
        """Run custom SQL query towards DB

        :param dataset: str.
            Name of dataset (collection)
        :param limit: int, default 10000.
            Number of datapoints (rows) to limit query to

        :return data: pd.DataFrame.
            Dataset, limited to last "limit" records

        Example
        ^^^^^^^
        .. code-block:: python

            >>> from factorymind.data_loader import FactoryDB

            >>> mydb = FactoryDB(apikey=YOUR-API-KEY)
            >>> query = \"\"\"
                        SELECT timestamp, sensor_00
                        FROM example_data.sensors
                        WHERE sensor_00 > 2.30;
                    \"\"\"
            >>> df = mydb.run_custom_query(query)
        """
        # self._check_if_dataset_exists(dataset)
        resp = requests.get(f"http://{self.api_url_base}/custom_query?query={query}")
        data_resp = self._get_response(resp)

        # Construct output
        if "columns" in data_resp.keys():
            data = pd.DataFrame(data_resp["data"], columns=data_resp["columns"])
        else:
            data = pd.DataFrame(data_resp["data"])

        return data

    def _check_if_dataset_exists(self, dataset):
        """Check if specified dataset name exists in DB"""
        if dataset not in self.datasets:
            raise DatasetDoesNotExistException(dataset)

    @staticmethod
    def _get_response(response):
        """Get response of API call"""
        if response.status_code != 200:
            # Something went wrong in API call
            raise CouldNotFetchDataException()
        else:
            data = response.json()
            return data



r"""
These are custom FM exception classes.

**References**

* Blog post\: `How to define custom exception classes in Python <https://towardsdatascience.com/how-to-define-custom-exception-classes-in-python-bfa346629bca>`_

Code documentation
^^^^^^^^^^^^^^^^^^
"""


class FactoryMindException(Exception):
    """Base class for other exceptions"""

    def __init__(self, msg="FMException"):
        self.message = msg
        super().__init__(self.message)


class CouldNotFetchDataException(FactoryMindException):
    """Raised when data cannot be fetched from FM database"""

    def __init__(self, msg="Could not fetch data from the FactoryMind DB"):
        self.message = msg
        super().__init__(self.message)

    def __str__(self):
        """Objects own message (self.message)"""
        if self.message:
            return "CouldNotFetchDataException, {} ".format(self.message)
        else:
            return "CouldNotFetchDataException has been raised"


class DatasetDoesNotExistException(FactoryMindException):
    """Raised when dataset (collection) does not exist in DB"""

    def __init__(
        self,
        dataset_name="",
        msg="",
    ):
        self.message = (
            msg
            + f"Could not find dataset in FactoryMind DB that matches the name '{dataset_name}'. Please check name of dataset."
        )
        super().__init__(self.message)



"""
Utility functions for timing (e.g. tic-yoc functionality)

Code documentation
^^^^^^^^^^^^^^^^^^
"""
import logging
import time

# Import standard modules
from contextlib import contextmanager


@contextmanager
def timeit_context(name: str, logger=None, space="  ", logging_level="debug", tic=None):
    """Time a piece of code

    Example
        ^^^^^^^
        .. code-block:: python

            >>> from factorymind.timing_utils import timeit_context

            >>> with timeit_context("Time to do some random list computation"):
                    l = [i ** 2 for i in range(10000000)]
                    del l
    """
    if logger is None:
        logger = logging.getLogger("root")
    tic = time.time() if tic is None else tic
    yield
    elapsedTime = time.time() - tic
    if logging_level == "info":
        # logging.basicConfig(level=logging.INFO)
        logger.info(space + "{}: {:.2f} sec".format(name, elapsedTime))
    elif logging_level == "debug":
        # logging.basicConfig(level=logging.DEBUG)
        logger.debug(space + "{}: {:.2f} sec".format(name, elapsedTime))


def setup_logger(name):
    r"""Set up custom logger

    **References**

    * Stackoverflow \: `Python logging module globally <https://stackoverflow.com/questions/7621897/python-logging-module-globally>`_
    """
    formatter = logging.Formatter(
        fmt="%(levelname)s:  %(asctime)s - %(module)s - %(message)s",
        datefmt="%Y/%m/%d %H:%M:%S",
    )

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    # logger.addHandler(handler)

    return logger



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['factorymind']

package_data = \
{'': ['*']}

install_requires = \
['bumpver>=2022.1116,<2023.0',
 'fastapi>=0.68.1,<0.69.0',
 'numpy>=1.21.2,<2.0.0',
 'pandas>=1.3.3,<2.0.0',
 'pytest>=6.2.5,<7.0.0',
 'python-dotenv>=0.19.0,<0.20.0',
 'requests>=2.26.0,<3.0.0',
 'uvicorn>=0.15.0,<0.16.0']

setup_kwargs = {
    'name': 'factorymind',
    'version': '0.1.3',
    'description': 'Python module for the FactoryMind platform',
    'long_description': "Documentation\n=============\nThis is the documentation for the Python package\n``factorymind`` which makes it easier to interact with the data\non the FactoryMind platform.\n\n.. image:: https://github.com/factorymind/factorymind/blob/master/docs/logo.png\n   :width: 200px\n\n\n************\nInstallation\n************\nFrom PyPi:\n\n.. code:: bash\n\n    pip install factorymind\n\nFrom source:\n\n.. code:: bash\n\n    <TODO: Document>\n\n***\nUse\n***\nTo interact with your data on the FactoryMind platform,\nrun\n\n.. code:: python\n\n    >>> from factorymind.data_loader import FactoryDB\n\n    >>> mydb = FactoryDB(apikey=YOUR-API-KEY)\n    >>> mydb.list_data_sources()\n\n    ['example_data.energy_demand', 'example_data.sensors', 'sensor_data.sensors', 'sensor_data.sensors_metadata']\n\nSee the `official docs <https://factorymind.readthedocs.io>`_ for more information and examples.\n",
    'author': 'FactoryMind',
    'author_email': 'enquiry@factorymind.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://factorymind.ai',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
